# Copyright 2022 PyQ Development Team

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import Any, Optional, Union

import torch

torch.set_default_dtype(torch.float64)

DEFAULT_MATRIX_DTYPE = torch.cdouble

IMAT = torch.eye(2, dtype=DEFAULT_MATRIX_DTYPE)
XMAT = torch.tensor([[0, 1], [1, 0]], dtype=DEFAULT_MATRIX_DTYPE)
YMAT = torch.tensor([[0, -1j], [1j, 0]], dtype=DEFAULT_MATRIX_DTYPE)
ZMAT = torch.tensor([[1, 0], [0, -1]], dtype=DEFAULT_MATRIX_DTYPE)
SMAT = torch.tensor([[1, 0], [0, 1j]], dtype=DEFAULT_MATRIX_DTYPE)
SDAGGERMAT = torch.tensor([[1, 0], [0, -1j]], dtype=DEFAULT_MATRIX_DTYPE)
TMAT = torch.tensor(
    [[1, 0], [0, torch.exp(torch.tensor(1.0j * torch.pi / 4))]], dtype=DEFAULT_MATRIX_DTYPE
)
NMAT = torch.tensor([[0, 0], [0, 1]], dtype=DEFAULT_MATRIX_DTYPE)
NDIAG = torch.tensor([0, 1], dtype=DEFAULT_MATRIX_DTYPE)
ZDIAG = torch.tensor([1, -1], dtype=DEFAULT_MATRIX_DTYPE)
IDIAG = torch.tensor([1, 1], dtype=DEFAULT_MATRIX_DTYPE)
SWAPMAT = torch.tensor(
    [[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]], dtype=DEFAULT_MATRIX_DTYPE
)
CSWAPMAT = torch.tensor(
    [
        [1, 0, 0, 0, 0, 0, 0, 0],
        [0, 1, 0, 0, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0, 0, 0],
        [0, 0, 0, 1, 0, 0, 0, 0],
        [0, 0, 0, 0, 1, 0, 0, 0],
        [0, 0, 0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 0, 0, 0, 1],
    ],
    dtype=DEFAULT_MATRIX_DTYPE,
)
HMAT = 1 / torch.sqrt(torch.tensor(2)) * torch.tensor([[1, 1], [1, -1]], dtype=DEFAULT_MATRIX_DTYPE)


OPERATIONS_DICT = {
    "I": IMAT,
    "X": XMAT,
    "Y": YMAT,
    "Z": ZMAT,
    "S": SMAT,
    "SDAGGER": SDAGGERMAT,
    "T": TMAT,
    "N": NMAT,
    "H": HMAT,
    "SWAP": SWAPMAT,
    "CSWAP": CSWAPMAT,
}


def ZZ(N: int, i: int = 0, j: int = 0, device: Union[str, torch.device] = "cpu") -> torch.Tensor:
    """
    Returns the tensor representation of the ZZ interaction operator
    between qubits i and j in a quantum circuit.

    Arguments:
        N (int): The total number of qubits in the circuit.
        i (int): Index of the first qubit (default: 0).
        j (int): Index of the second qubit (default: 0).
        device (Union[str, torch.device]): Device to store the tensor on (default: "cpu").

    Returns:
        torch.Tensor: The tensor representation of the ZZ interaction operator.

    Examples:
    ```python exec="on" source="above" result="json"
    from pyqtorch.matrices import ZZ
    result=ZZ(2, 0, 1)
    print(result) #tensor([ 1.+0.j, -1.+0.j, -1.+0.j,  1.-0.j], dtype=torch.complex128)
    ```
    """
    if i == j:
        return torch.ones(2**N).to(device)

    op_list = [ZDIAG.to(device) if k in [i, j] else IDIAG.to(device) for k in range(N)]
    operator = op_list[0]
    for op in op_list[1::]:
        operator = torch.kron(operator, op)

    return operator


def NN(N: int, i: int = 0, j: int = 0, device: Union[str, torch.device] = "cpu") -> torch.Tensor:
    """
    Returns the tensor representation of the NN interaction operator
    between qubits i and j in a quantum circuit.

    Arguments:
        N (int): The total number of qubits in the circuit.
        i (int): Index of the first qubit (default: 0).
        j (int): Index of the second qubit (default: 0).
        device (Union[str, torch.device]): Device to store the tensor on (default: "cpu").

    Returns:
        torch.Tensor: The tensor representation of the NN interaction operator.

    Examples:
    ```python exec="on" source="above" result="json"
    from pyqtorch.matrices import NN
    result=NN(2, 0, 1)
    print(result) #tensor([0.+0.j, 0.+0.j, 0.+0.j, 1.+0.j], dtype=torch.complex128)
    ```
    """
    if i == j:
        return torch.ones(2**N, dtype=DEFAULT_MATRIX_DTYPE).to(device)

    op_list = [NDIAG.to(device) if k in [i, j] else IDIAG.to(device) for k in range(N)]
    operator = op_list[0]
    for op in op_list[1::]:
        operator = torch.kron(operator, op)

    return operator


def single_Z(N: int, i: int = 0, device: Union[str, torch.device] = "cpu") -> torch.Tensor:
    op_list = [ZDIAG.to(device) if k == i else IDIAG.to(device) for k in range(N)]
    operator = op_list[0]
    for op in op_list[1::]:
        operator = torch.kron(operator, op)

    return operator


def single_N(N: int, i: int = 0, device: Union[str, torch.device] = "cpu") -> torch.Tensor:
    op_list = [NDIAG.to(device) if k == i else IDIAG.to(device) for k in range(N)]
    operator = op_list[0]
    for op in op_list[1::]:
        operator = torch.kron(operator, op)

    return operator


def sum_Z(N: int, device: Union[str, torch.device] = "cpu") -> torch.Tensor:
    H = torch.zeros(2**N, dtype=torch.cdouble).to(device)
    for i in range(N):
        H += single_Z(N, i, device)
    return H


def sum_N(N: int, device: Union[str, torch.device] = "cpu") -> torch.Tensor:
    H = torch.zeros(2**N, dtype=torch.cdouble).to(device)
    for i in range(N):
        H += single_N(N, i, device)
    return H


def generate_ising_from_graph(
    graph: Any,  # optional library type
    precomputed_zz: Optional[torch.Tensor] = None,
    type_ising: str = "Z",
    device: Union[str, torch.device] = "cpu",
) -> torch.Tensor:
    N = graph.number_of_nodes()
    # construct the hamiltonian
    H = torch.zeros(2**N, dtype=DEFAULT_MATRIX_DTYPE).to(device)

    for edge in graph.edges.data():
        if precomputed_zz is not None:
            if (edge[0], edge[1]) in precomputed_zz[N]:
                key = (edge[0], edge[1])
            else:
                key = (edge[1], edge[0])
            H += precomputed_zz[N][key]
        else:
            if type_ising == "Z":
                H += ZZ(N, edge[0], edge[1], device)
            elif type_ising == "N":
                H += NN(N, edge[0], edge[1], device)
            else:
                raise ValueError("'type_ising' must be in ['Z', 'N']")

    return H



from __future__ import annotations

from enum import Enum

import torch
from numpy.typing import ArrayLike

from pyqtorch.core.batched_operation import (
    create_controlled_batch_from_operation,
)
from pyqtorch.core.utils import _apply_batch_gate
from pyqtorch.matrices import DEFAULT_MATRIX_DTYPE, OPERATIONS_DICT
from pyqtorch.modules.abstract import AbstractGate
from pyqtorch.modules.ops import _vmap_gate
from pyqtorch.modules.utils import rot_matrices

torch.set_default_dtype(torch.float64)


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))  # type: ignore


class ApplyFn(StrEnum):
    """Which function to use to perform matmul between gate and state."""

    VMAP = "vmap"
    EINSUM = "einsum"


APPLY_FN_DICT = {ApplyFn.VMAP: _vmap_gate, ApplyFn.EINSUM: _apply_batch_gate}
DEFAULT_APPLY_FN = ApplyFn.EINSUM


class RotationGate(AbstractGate):
    n_params = 1

    def __init__(
        self, gate: str, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN
    ):
        super().__init__(qubits, n_qubits)
        self.gate = gate
        self.register_buffer("imat", OPERATIONS_DICT["I"])
        self.register_buffer("paulimat", OPERATIONS_DICT[gate])
        self.apply_fn = APPLY_FN_DICT[apply_fn_type]

    def matrices(self, thetas: torch.Tensor) -> torch.Tensor:
        theta = thetas.squeeze(0) if thetas.ndim == 2 else thetas
        batch_size = len(theta)
        return rot_matrices(theta, self.paulimat, self.imat, batch_size)

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = matrices.size(-1)
        return self.apply_fn(state, matrices, self.qubits, self.n_qubits, batch_size)

    def forward(self, state: torch.Tensor, thetas: torch.Tensor) -> torch.Tensor:
        mats = self.matrices(thetas)
        return self.apply(mats, state)

    def extra_repr(self) -> str:
        return f"qubits={self.qubits}, n_qubits={self.n_qubits}"


class U(AbstractGate):
    n_params = 3

    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a parametrized arbitrary rotation along the axes of the Bloch sphere.

        The angles `phi, theta, omega` in tensor format, applied as:

            U(phi, theta, omega) = RZ(omega)RY(theta)RZ(phi)

        Arguments:
            qubits (ArrayLike): The target qubits for the U gate. It should be a list of qubits.
            n_qubits (int): The total number of qubits in the circuit.

        """

        super().__init__(qubits, n_qubits)

        self.register_buffer(
            "a", torch.tensor([[1, 0], [0, 0]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2)
        )
        self.register_buffer(
            "b", torch.tensor([[0, 1], [0, 0]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2)
        )
        self.register_buffer(
            "c", torch.tensor([[0, 0], [1, 0]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2)
        )
        self.register_buffer(
            "d", torch.tensor([[0, 0], [0, 1]], dtype=DEFAULT_MATRIX_DTYPE).unsqueeze(2)
        )

    def matrices(self, thetas: torch.Tensor) -> torch.Tensor:
        if thetas.ndim == 1:
            thetas = thetas.unsqueeze(1)
        assert thetas.size(0) == 3
        phi, theta, omega = thetas[0, :], thetas[1, :], thetas[2, :]
        batch_size = thetas.size(1)

        t_plus = torch.exp(-1j * (phi + omega) / 2)
        t_minus = torch.exp(-1j * (phi - omega) / 2)
        sin_t = torch.sin(theta / 2).unsqueeze(0).unsqueeze(1).repeat((2, 2, 1))
        cos_t = torch.cos(theta / 2).unsqueeze(0).unsqueeze(1).repeat((2, 2, 1))

        a = self.a.repeat(1, 1, batch_size) * cos_t * t_plus
        b = self.b.repeat(1, 1, batch_size) * sin_t * torch.conj(t_minus)
        c = self.c.repeat(1, 1, batch_size) * sin_t * t_minus
        d = self.d.repeat(1, 1, batch_size) * cos_t * torch.conj(t_plus)
        return a - b + c + d

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = matrices.size(-1)
        return _apply_batch_gate(state, matrices, self.qubits, self.n_qubits, batch_size)

    def forward(self, state: torch.Tensor, thetas: torch.Tensor) -> torch.Tensor:
        """
        Represents batched state and theta forwarding  to the matrices
        and apply functions that are part of the U gate implementation

        Arguments:
            state  (torch.Tensor): batched state
            thetas (torch.Tensor): Tensor of size 3 ,contains values of `phi`/`theta`/`omega`

        Returns:
            torch.Tensor: the resulting state after applying the gate
        """
        mats = self.matrices(thetas)
        return self.apply(mats, state)


class ControlledRotationGate(AbstractGate):
    n_params = 1

    def __init__(
        self, gate: str, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN
    ):
        super().__init__(qubits, n_qubits)
        self.gate = gate
        self.register_buffer("imat", OPERATIONS_DICT["I"])
        self.register_buffer("paulimat", OPERATIONS_DICT[gate])
        self.apply_fn = APPLY_FN_DICT[apply_fn_type]

    def matrices(self, thetas: torch.Tensor) -> torch.Tensor:
        theta = thetas.squeeze(0) if thetas.ndim == 2 else thetas
        batch_size = len(theta)
        return rot_matrices(theta, self.paulimat, self.imat, batch_size)

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = matrices.size(-1)
        controlled_mats = create_controlled_batch_from_operation(
            matrices, batch_size, len(self.qubits) - 1
        )
        return self.apply_fn(state, controlled_mats, self.qubits, self.n_qubits, batch_size)

    def forward(self, state: torch.Tensor, thetas: torch.Tensor) -> torch.Tensor:
        mats = self.matrices(thetas)
        return self.apply(mats, state)


class RX(RotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents an X-axis rotation (RX) gate in a quantum circuit.
        The RX gate class creates a single-qubit RX gate that performs
        a given rotation around the X axis.

        Arguments:
            qubits (ArrayLike):The list of qubits the controlled RX gate acts on.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an RX gate
        rx_gate = pyq.RX(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        #Create a random theta angle
        theta = torch.rand(1)

        # Every rotational gate accepts a second parameter that is expected to be a Theta angle.
        # Apply the RX gate to the zero state with your random theta as a second parameter.

        result=rx_gate(z_state, theta)
        print(result)

        ```
        """
        super().__init__("X", qubits, n_qubits, apply_fn_type)


class RY(RotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a Y-axis rotation (RY) gate in a quantum circuit.
        The RY gate class creates a single-qubit RY gate that performs
        a given rotation around the Y axis.

        Arguments:
            qubits (ArrayLike): The qubit index to apply the RY gate to.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an RY gate
        ry_gate = pyq.RY(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Create a random theta angle
        theta = torch.rand(1)

        # Apply the RY gate to the zero state with the random theta angle
        result = ry_gate(z_state, theta)
        print(result)
        ```
        """
        super().__init__("Y", qubits, n_qubits, apply_fn_type)


class RZ(RotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a Z-axis rotation (RZ) gate in a quantum circuit.
        The RZ gate class creates a single-qubit RZ gate that performs
        a given rotation around the Z axis.


        Arguments:
            qubits (ArrayLike): The qubit index or list of qubit indices to apply the RZ gate to.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an RZ gate
        rz_gate = pyq.RZ(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Create a random theta angle
        theta = torch.rand(1)

        # Apply the RZ gate to the zero state with the random theta angle
        result = rz_gate(z_state, theta)
        print(result)
        ```
        """
        super().__init__("Z", qubits, n_qubits, apply_fn_type)


class PHASE(RotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a PHASE rotation gate in a quantum circuit.

        Arguments:
            qubits (ArrayLike): The qubit index or list of qubit indices.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an PHASE gate
        gate = pyq.PHASE(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Create a random theta angle
        theta = torch.rand(1)

        # Apply the PHASE gate to the zero state with the random theta angle
        result = gate(z_state, theta)
        print(result)
        ```
        """
        super().__init__("I", qubits, n_qubits)
        self.apply_fn = APPLY_FN_DICT[apply_fn_type]

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = matrices.size(-1)
        return self.apply_fn(state, matrices, self.qubits, self.n_qubits, batch_size)

    def forward(self, state: torch.Tensor, thetas: torch.Tensor) -> torch.Tensor:
        mats = self.matrices(thetas)
        return self.apply(mats, state)

    def matrices(self, thetas: torch.Tensor) -> torch.Tensor:
        theta = thetas.squeeze(0) if thetas.ndim == 2 else thetas
        batch_mat = self.imat.unsqueeze(2).repeat(1, 1, len(theta))
        batch_mat[1, 1, :] = torch.exp(1.0j * thetas).unsqueeze(0).unsqueeze(1)
        return batch_mat


class CRX(ControlledRotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a controlled-X-axis rotation (CRX) gate in a quantum circuit.
        The CRX gate class creates a controlled RX gate, applying the RX according
        to the control qubit state.


        Arguments:
            qubits (ArrayLike):The list of qubits the controlled CRX gate acts on.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq
        # Create a CRX gate
        #The CRX gate is a controlled version of the RX gate.
        #It applies an RX rotation to the target qubit based on the state of the control qubit.
        #The gate takes two qubits as input: the control qubit and the target qubit.

        crx_gate = pyq.CRX(qubits=[0, 1], n_qubits=2)

        # Create a X gate
        x_gate=pyq.X(qubits=[0], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        #Apply an X gate to zero state to change its  state from 0 to 1
        activation_state=x_gate(z_state)

        # Create a random theta angle
        theta = torch.rand(1)

        # Apply the CRX gate to the activation state with the random theta angle
        result = crx_gate(activation_state, theta)
        print(result)

        ```
        """
        super().__init__("X", qubits, n_qubits, apply_fn_type)


class CRY(ControlledRotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a controlled-Y-axis rotation (CRY) gate in a quantum circuit.
        The CRY gate class creates a controlled RY gate, applying the RY according
        to the control qubit state.


        Arguments:
            qubits (ArrayLike):The list of qubits the controlled CRY gate acts on.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a CRY gate
        # The CRY gate is a controlled version of the RY gate.
        # It applies an RY rotation to the target qubit based on the state of the control qubit.
        # The gate takes two qubits as input: the control qubit and the target qubit.
        cry_gate = pyq.CRY(qubits=[0, 1], n_qubits=2)

        # Create a Y gate
        y_gate = pyq.Y(qubits=[0], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        # Apply a Y gate to the zero state to change its state
        activation_state = y_gate(z_state)

        # Create a random theta angle
        theta = torch.rand(1)

        # Apply the CRY gate to the activation state with the random theta angle
        result = cry_gate(activation_state, theta)
        print(result)
        ```
        """
        super().__init__("Y", qubits, n_qubits, apply_fn_type)


class CRZ(ControlledRotationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a controlled-Z-axis rotation (CRZ) gate in a quantum circuit.
        The CRZ gate class creates a controlled RZ gate, applying the RZ according
        to the control qubit state.

        Arguments:
            qubits (ArrayLike):The list of qubits the controlled CRZ gate acts on.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a CRZ gate
        # The CRZ gate is a controlled version of the RZ gate.
        # It applies an RZ rotation to the target qubit based on the state of the control qubit.
        # The gate takes two qubits as input: the control qubit and the target qubit.
        crz_gate = pyq.CRZ(qubits=[0, 1], n_qubits=2)

        # Create a X gate
        x_gate = pyq.X(qubits=[0], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        # Apply a X gate to the zero state to change its state
        activation_state = x_gate(z_state)

        # Create a random theta angle
        theta = torch.rand(1)

        # Apply the CRZ gate to the activation state with the random theta angle
        result = crz_gate(activation_state, theta)
        print(result)
        ```
        """
        super().__init__("Z", qubits, n_qubits, apply_fn_type)


class CPHASE(AbstractGate):
    n_params = 1

    def __init__(self, qubits: ArrayLike, n_qubits: int, apply_fn_type: ApplyFn = DEFAULT_APPLY_FN):
        """
        Represents a controlled-phase (CPHASE) gate in a quantum circuit.
        The CPhase gate class creates a controlled Phase gate, applying the PhaseGate
        according to the control qubit state.

        Arguments:
            qubits (ArrayLike): The control and target qubits for the CPHASE gate.
            n_qubits (int): The total number of qubits in the circuit.

        """

        super().__init__(qubits, n_qubits)

        self.register_buffer("imat", torch.eye(2 ** len(qubits), dtype=DEFAULT_MATRIX_DTYPE))
        self.apply_fn = APPLY_FN_DICT[apply_fn_type]

    def matrices(self, thetas: torch.Tensor) -> torch.Tensor:
        theta = thetas.squeeze(0) if thetas.ndim == 2 else thetas
        batch_size = len(theta)
        mat = self.imat.repeat((batch_size, 1, 1))
        mat = torch.permute(mat, (1, 2, 0))
        phase_rotation_angles = torch.exp(torch.tensor(1j) * theta).unsqueeze(0).unsqueeze(1)
        mat[-1, -1, :] = phase_rotation_angles
        return mat

    def apply(self, matrices: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        batch_size = matrices.size(-1)
        return self.apply_fn(state, matrices, self.qubits, self.n_qubits, batch_size)

    def forward(self, state: torch.Tensor, thetas: torch.Tensor) -> torch.Tensor:
        mats = self.matrices(thetas)
        return self.apply(mats, state)



from __future__ import annotations

from functools import reduce
from typing import Tuple

import torch
from numpy import log2

from pyqtorch.core.utils import _apply_gate
from pyqtorch.matrices import DEFAULT_MATRIX_DTYPE
from pyqtorch.modules.abstract import AbstractGate


def normalize(wf: torch.Tensor) -> torch.Tensor:
    return wf / torch.sqrt((wf.abs() ** 2).sum())


def is_normalized(state: torch.Tensor, atol: float = 1e-15) -> bool:
    n_qubits = len(state.size()) - 1
    batch_size = state.size()[-1]
    state = state.reshape((2**n_qubits, batch_size))
    sum_probs = (state.abs() ** 2).sum(dim=0)
    ones = torch.ones(batch_size)
    return torch.allclose(sum_probs, ones, rtol=0.0, atol=atol)  # type:ignore[no-any-return]


def is_diag(H: torch.Tensor) -> bool:
    """
    Returns True if Hamiltonian H is diagonal.
    """
    return len(torch.abs(torch.triu(H, diagonal=1)).to_sparse().coalesce().values()) == 0


def is_real(H: torch.Tensor) -> bool:
    """
    Returns True if Hamiltonian H is real.
    """
    return len(torch.imag(H).to_sparse().coalesce().values()) == 0


def overlap(state: torch.Tensor, other: torch.Tensor) -> torch.Tensor:
    n_qubits = len(state.size()) - 1
    batch_size = state.size()[-1]
    state = state.reshape((2**n_qubits, batch_size))
    other = other.reshape((2**n_qubits, batch_size))
    res = []
    for i in range(batch_size):
        ovrlp = torch.real(torch.sum(torch.conj(state[:, i]) * other[:, i]))
        res.append(ovrlp)
    return torch.stack(res)


def rot_matrices(
    theta: torch.Tensor, P: torch.Tensor, I: torch.Tensor, batch_size: int  # noqa: E741
) -> torch.Tensor:
    """
    Returns:
        torch.Tensor: a batch of gates after applying theta
    """
    cos_t = torch.cos(theta / 2).unsqueeze(0).unsqueeze(1)
    cos_t = cos_t.repeat((2, 2, 1))
    sin_t = torch.sin(theta / 2).unsqueeze(0).unsqueeze(1)
    sin_t = sin_t.repeat((2, 2, 1))

    batch_imat = I.unsqueeze(2).repeat(1, 1, batch_size)
    batch_operation_mat = P.unsqueeze(2).repeat(1, 1, batch_size)

    return cos_t * batch_imat - 1j * sin_t * batch_operation_mat


def zero_state(
    n_qubits: int,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
) -> torch.Tensor:
    """
    Generates the zero state for a specified number of qubits.

    Arguments:
        n_qubits (int): The number of qubits for which the zero state is to be generated.
        batch_size (int): The batch size for the zero state.
        device (str): The device on which the zero state tensor is to be allocated eg cpu or gpu.
        dtype (DEFAULT_MATRIX_DTYPE): The data type of the zero state tensor.

    Returns:
        torch.Tensor: A tensor representing the zero state.
        The shape of the tensor is (batch_size, 2^n_qubits),
        where 2^n_qubits is the total number of possible states for the given number of qubits.
        The data type of the tensor is specified by the dtype parameter.

    Examples:
    ```python exec="on" source="above" result="json"
    import torch
    import pyqtorch.modules as pyq

    state = pyq.zero_state(n_qubits=2)
    print(state)  #tensor([[[1.+0.j],[0.+0.j]],[[0.+0.j],[0.+0.j]]], dtype=torch.complex128)
    ```
    """
    state = torch.zeros((2**n_qubits, batch_size), dtype=dtype, device=device)
    state[0] = 1
    state = state.reshape([2] * n_qubits + [batch_size])
    return state


def uniform_state(
    n_qubits: int,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
) -> torch.Tensor:
    """
    Generates the uniform state for a specified number of qubits.
    Returns a tensor representing the uniform state.
    The shape of the tensor is (2^n_qubits, batch_size),
    where 2^n_qubits is the total number of possible states for the given number of qubits.
    The data type of the tensor is specified by the dtype parameter.
    Each element of the tensor is initialized to 1/sqrt(2^n_qubits),
    ensuring that the total probability of the state is equal to 1.

    Arguments:
        n_qubits (int): The number of qubits for which the uniform state is to be generated.
        batch_size (int): The batch size for the uniform state.
        device (str): The device on which the uniform state tensor is to be allocated.
        dtype (DEFAULT_MATRIX_DTYPE): The data type of the uniform state tensor.

    Returns:
        torch.Tensor: A tensor representing the uniform state.


    Examples:
    ```python exec="on" source="above" result="json"
    import torch
    import pyqtorch.modules as pyq

    state = pyq.uniform_state(n_qubits=2)
    print(state)
    #tensor([[[0.5000+0.j],[0.5000+0.j]],[[0.5000+0.j],[0.5000+0.j]]], dtype=torch.complex128)
    ```
    """
    state = torch.ones((2**n_qubits, batch_size), dtype=dtype, device=device)
    state = state / torch.sqrt(torch.tensor(2**n_qubits))
    state = state.reshape([2] * n_qubits + [batch_size])
    return state


def random_state(
    n_qubits: int,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
) -> torch.Tensor:
    def _rand(n_qubits: int) -> torch.Tensor:
        N = 2**n_qubits
        x = -torch.log(torch.rand(N))
        sumx = torch.sum(x)
        phases = torch.rand(N) * 2.0 * torch.pi
        return normalize(
            (torch.sqrt(x / sumx) * torch.exp(1j * phases)).reshape(N, 1).type(dtype).to(device)
        )

    _state = torch.concat(tuple(_rand(n_qubits) for _ in range(batch_size)), dim=1)
    return _state.reshape([2] * n_qubits + [batch_size])


def flatten_wf(wf: torch.Tensor) -> torch.Tensor:
    return torch.flatten(wf, start_dim=0, end_dim=-2).t()


def invert_endianness(wf: torch.Tensor) -> torch.Tensor:
    """
    Inverts the endianness of a wave function.

    Args:
        wf (Tensor): the target wf as a torch Tensor of shape batch_size X 2**n_qubits

    Returns:
        The inverted wave function.
    """
    try:
        wf = flatten_wf(wf)
    except RuntimeError:
        wf = wf
    n_qubits = int(log2(wf.shape[1]))
    ls = list(range(2**n_qubits))
    permute_ind = torch.tensor([int(f"{num:0{n_qubits}b}"[::-1], 2) for num in ls])
    return wf[:, permute_ind]


def _apply_parallel(
    state: torch.Tensor,
    thetas: torch.Tensor,
    gates: Tuple[AbstractGate] | list[AbstractGate],
    n_qubits: int,
) -> torch.Tensor:
    qubits_list: list[Tuple] = [g.qubits for g in gates]
    mats = [g.matrices(thetas) for g in gates]

    return reduce(
        lambda state, inputs: _apply_gate(state, *inputs, N_qubits=n_qubits),  # type: ignore
        zip(mats, qubits_list),
        state,
    )



from __future__ import annotations

from enum import Enum
from functools import lru_cache
from typing import Any, Optional, Tuple, Union

import torch
from torch.nn import Module

from pyqtorch.core.utils import _apply_batch_gate
from pyqtorch.modules.utils import is_diag, is_real

BATCH_DIM = 2


class HamEvo(torch.nn.Module):
    """
    Base class for Hamiltonian evolution classes, performing the evolution using RK4 method.

    Args:
        H (tensor): Hamiltonian tensor.
        t (tensor): Time tensor.
        qubits (Any): Qubits for operation.
        n_qubits (int): Number of qubits.
        n_steps (int): Number of steps to be performed in RK4-based evolution. Defaults to 100.


    """

    def __init__(
        self, H: torch.Tensor, t: torch.Tensor, qubits: Any, n_qubits: int, n_steps: int = 100
    ):
        super().__init__()
        self.H: torch.Tensor
        self.t: torch.Tensor
        self.qubits = qubits
        self.n_qubits = n_qubits
        self.n_steps = n_steps
        if H.ndim == 2:
            H = H.unsqueeze(2)
        if H.size(-1) == t.size(0) or t.size(0) == 1:
            self.register_buffer("H", H)
            self.register_buffer("t", t)
        elif H.size(-1) == 1:
            (x, y, _) = H.size()
            self.register_buffer("H", H.expand(x, y, t.size(0)))
            self.register_buffer("t", t)
        else:
            msg = "H and t batchsizes either have to match or (one of them has to) be equal to one."
            raise ValueError(msg)

    def apply(self, state: torch.Tensor) -> torch.Tensor:
        """
        Applies the Hamiltonian evolution operation on the given state using RK4 method.

        Args:
            state (tensor): Input quantum state.

        Returns:
            tensor: Output state after Hamiltonian evolution.
        """

        batch_size = max(state.size(-1), self.H.size(-1))
        if state.size(-1) == 1:
            state = state.repeat(*[1 for _ in range(len(state.size()) - 1)], batch_size)

        h = self.t.reshape((1, -1)) / self.n_steps
        for _ in range(self.n_qubits - 1):
            h = h.unsqueeze(0)

        h = h.expand_as(state)
        _state = state.clone()
        for _ in range(self.n_steps):
            k1 = -1j * _apply_batch_gate(_state, self.H, self.qubits, self.n_qubits, batch_size)
            k2 = -1j * _apply_batch_gate(
                _state + h / 2 * k1, self.H, self.qubits, self.n_qubits, batch_size
            )
            k3 = -1j * _apply_batch_gate(
                _state + h / 2 * k2, self.H, self.qubits, self.n_qubits, batch_size
            )
            k4 = -1j * _apply_batch_gate(
                _state + h * k3, self.H, self.qubits, self.n_qubits, batch_size
            )
            _state += h / 6 * (k1 + 2 * k2 + 2 * k3 + k4)

        return _state

    def forward(self, state: torch.Tensor) -> torch.Tensor:
        """
        Forward pass of the module, applies the Hamiltonian evolution operation.

        Args:
            state (tensor): Input quantum state.

        Returns:
            tensor: Output state after Hamiltonian evolution.
        """

        return self.apply(state)


@lru_cache(maxsize=256)
def diagonalize(H: torch.Tensor) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """
    Diagonalizes an Hermitian Hamiltonian, returning eigenvalues and eigenvectors.
    First checks if it's already diagonal, and second checks if H is real.
    """

    if is_diag(H):
        # Skips diagonalization
        eig_values = torch.diagonal(H)
        eig_vectors = None
    else:
        if is_real(H):
            eig_values, eig_vectors = torch.linalg.eigh(H.real)
            eig_values = eig_values.to(torch.cdouble)
            eig_vectors = eig_vectors.to(torch.cdouble)
        else:
            eig_values, eig_vectors = torch.linalg.eigh(H)

    return eig_values, eig_vectors


class HamEvoEig(HamEvo):
    """
    Class for Hamiltonian evolution operation using Eigenvalue Decomposition method.

    Args:
        H (tensor): Hamiltonian tensor
        t (tensor): Time tensor
        qubits (Any): Qubits for operation
        n_qubits (int): Number of qubits
        n_steps (int): Number of steps to be performed, defaults to 100
    """

    def __init__(
        self, H: torch.Tensor, t: torch.Tensor, qubits: Any, n_qubits: int, n_steps: int = 100
    ):
        super().__init__(H, t, qubits, n_qubits, n_steps)
        if len(self.H.size()) < 3:
            self.H = self.H.unsqueeze(2)
        batch_size_h = self.H.size()[BATCH_DIM]
        batch_size_t = self.t.size(0)

        self._eigs = []
        if batch_size_h == batch_size_t or batch_size_t == 1:
            for i in range(batch_size_h):
                eig_values, eig_vectors = diagonalize(self.H[..., i])
                self._eigs.append((eig_values, eig_vectors))
        elif batch_size_h == 1:
            eig_values, eig_vectors = diagonalize(self.H[..., 0])
            for i in range(batch_size_t):
                self._eigs.append((eig_values, eig_vectors))
        else:
            msg = "H and t batchsizes either have to match or (one of them has to) be equal to one."
            raise ValueError(msg)
        self.batch_size = max(batch_size_h, batch_size_t)

    def apply(self, state: torch.Tensor) -> torch.Tensor:
        """
        Applies the Hamiltonian evolution operation on the given state
        using Eigenvalue Decomposition method.

        Args:
            state (tensor): Input quantum state.

        Returns:
            tensor: Output state after Hamiltonian evolution.
        """

        (x, y, _) = self.H.size()
        evol_operator = torch.zeros(x, y, self.batch_size).to(torch.cdouble)
        t_evo = self.t.repeat(self.batch_size) if self.t.size(0) == 1 else self.t

        for i, (eig_values, eig_vectors) in enumerate(self._eigs):
            if eig_vectors is None:
                # Compute e^(-i H t)
                evol_operator[..., i] = torch.diag(torch.exp(-1j * eig_values * t_evo[i]))

            else:
                # Compute e^(-i D t)
                eig_exp = torch.diag(torch.exp(-1j * eig_values * t_evo[i]))
                # e^(-i H t) = V.e^(-i D t).V^\dagger
                evol_operator[..., i] = torch.matmul(
                    torch.matmul(eig_vectors, eig_exp),
                    torch.conj(eig_vectors.transpose(0, 1)),
                )

        return _apply_batch_gate(state, evol_operator, self.qubits, self.n_qubits, self.batch_size)


class HamEvoExp(HamEvo):
    """
    Class for Hamiltonian evolution operation, using matrix exponential method.

    Args:
        H (tensor): Hamiltonian tensor
        t (tensor): Time tensor
        qubits (Any): Qubits for operation
        n_qubits (int): Number of qubits
        n_steps (int): Number of steps to be performed, defaults to 100.
    """

    def __init__(
        self, H: torch.Tensor, t: torch.Tensor, qubits: Any, n_qubits: int, n_steps: int = 100
    ):
        super().__init__(H, t, qubits, n_qubits, n_steps)
        if len(self.H.size()) < 3:
            self.H = self.H.unsqueeze(2)
        batch_size_h = self.H.size()[BATCH_DIM]

        # Check if all hamiltonians in the batch are diagonal
        diag_check = torch.tensor([is_diag(self.H[..., i]) for i in range(batch_size_h)])
        self.batch_is_diag = bool(torch.prod(diag_check))

    def apply(self, state: torch.Tensor) -> torch.Tensor:
        """
        Applies the Hamiltonian evolution operation
        on the given state using matrix exponential method.

        Args:
            state (tensor): Input quantum state.

        Returns:
            tensor: Output state after Hamiltonian evolution.
        """

        batch_size_t = len(self.t)
        batch_size_h = self.H.size()[BATCH_DIM]
        t_evo = self.t

        if self.batch_is_diag:
            # Skips the matrix exponential for diagonal hamiltonians
            H_diagonals = torch.diagonal(self.H)
            evol_exp_arg = H_diagonals * (-1j * t_evo).view((-1, 1))
            evol_operator_T = torch.diag_embed(torch.exp(evol_exp_arg))
            evol_operator = torch.transpose(evol_operator_T, 0, -1)
        else:
            H_T = torch.transpose(self.H, 0, -1)
            evol_exp_arg = H_T * (-1j * t_evo).view((-1, 1, 1))
            evol_operator_T = torch.linalg.matrix_exp(evol_exp_arg)
            evol_operator = torch.transpose(evol_operator_T, 0, -1)

        batch_size = max(batch_size_h, batch_size_t)
        return _apply_batch_gate(state, evol_operator, self.qubits, self.n_qubits, batch_size)


class HamEvoType(Enum):
    """
    An Enumeration to represent types of Hamiltonian Evolution

    RK4: Hamiltonian evolution performed using the 4th order Runge-Kutta method.
    EIG: Hamiltonian evolution performed using Eigenvalue Decomposition.
    EXP: Hamiltonian evolution performed using the Exponential of the Hamiltonian.
    """

    RK4 = HamEvo
    EIG = HamEvoEig
    EXP = HamEvoExp


class HamiltonianEvolution(Module):
    """
    A module to encapsulate Hamiltonian Evolution operations.

    Performs Hamiltonian Evolution using different strategies
    such as, RK4, Eigenvalue Decomposition, and Exponential,
    Based on the 'hamevo_type' parameter.

    Attributes:
        qubits: A list of qubits to be used in the operation
        n_qubits (int): Total number of qubits
        n_steps  (int): The number of steps to be performed. Defaults to 100
        hamevo_type (Enum or str): The type of Hamiltonian evolution to be performed.
                     Must be a member of the `HamEvoType` enum or equivalent string.
                     Defaults to HamEvoExp.

    Examples:
    (1)
    # Instantiate HamiltonianEvolution with RK4 string input
        >>> hamiltonian_evolution = HamiltonianEvolution(qubits, n_qubits, 100, "RK4")
    # Use the HamiltonianEvolution instance to evolve the state
        >>> output_state = hamiltonian_evolution(H, t, state)

    (2)
    # Instantiate HamiltonianEvolution with HamEvoType. input
        >>>H_evol = HamiltonianEvolution(qubits, n_qubits, 100, HamEvoType.Eig)
    # Use the HamiltonianEvolution instance to evolve the state
        >>>output = H_evol(H, t, state) # Use the HamiltonianEvolution instance to evolve the state

    """

    def __init__(
        self,
        qubits: Any,
        n_qubits: int,
        n_steps: int = 100,
        hamevo_type: Union[HamEvoType, str] = HamEvoType.EXP,
    ):
        super().__init__()
        self.qubits = qubits
        self.n_qubits = n_qubits
        self.n_steps = n_steps

        # Handles the case where the Hamiltonian Evolution type is provided as a string
        if isinstance(hamevo_type, str):
            try:
                hamevo_type = HamEvoType[hamevo_type.upper()]
            except KeyError:
                allowed = [e.name for e in HamEvoType]
                raise ValueError(
                    f"Invalid Hamiltonian Evolution type: {hamevo_type}. Expected from: {allowed}"
                )

        self.hamevo_type = hamevo_type

    def get_hamevo_instance(self, H: torch.Tensor, t: torch.Tensor) -> torch.nn.Module:
        """
        Returns an instance of the Hamiltonian evolution object of the appropriate type.

        Args:
            H (tensor): The Hamiltonian to be used in the evolution.
            t (tensor): The evolution time.

        Returns:
            An instance of the Hamiltonian evolution object of the appropriate type.

        """

        return self.hamevo_type.value(H, t, self.qubits, self.n_qubits, self.n_steps)

    def forward(self, H: torch.Tensor, t: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        """
        Performs Hamiltonian evolution on the given state.

        Args:
            H (tensor): The Hamiltonian to be used in the evolution.
            t (tensor): The evolution time.
            state (tensor): The state on which to perform Hamiltonian evolution.

        Returns:
            The state (tensor) after Hamiltonian evolution.
        """
        ham_evo_instance = self.get_hamevo_instance(H, t)
        return ham_evo_instance.forward(state)



import torch
from memory_profiler import profile
import pyqtorch.modules as pyq
from pyqtorch.core.utils import _apply_gate, _apply_batch_gate
from timeit import timeit
import time
from itertools import product


batch_size = 1
N_QUBITS = 25
thetas = torch.rand(batch_size)
QUBIT = 0


# rxsparsemat = _fill_identities(rxmat,(0,), [i for i in range(10)]).squeeze(0).to_sparse()

rx = pyq.RX([0], N_QUBITS)



def run_mem_script():
    state = pyq.zero_state(N_QUBITS, batch_size)
    for i in range(2):
        state = _apply_gate(state, rx.matrices(thetas).squeeze(-1), rx.qubits, rx.n_qubits)


# @timeit
def run_time_script():
    _vmap_gate(state, rx.matrices(thetas), rx.qubits, rx.n_qubits,batch_size)


@profile
def matmul():

    pass


def find_indices_of_bitstrings_with_zero_at_i(N:int, i:int, bit:str ='0'):
    if i >= N or i < 0:
        raise ValueError("Index i must be in the range [0, N-1]")
    
    indices = []
    for idx, p in enumerate(product('01', repeat=N)):
        if p[i] == bit:
            indices.append(idx)
            
    return indices

# def permute_matmul() -> torch.Tensor:



if __name__ == '__main__':
    start = time.time()
    indices = find_indices_of_bitstrings_with_zero_at_i(4, 0)
    end = time.time()
    print(indices)
    print(abs(end-start))
    # run_time()


from __future__ import annotations

import torch


def _vmap_gate(
    state: torch.Tensor,
    mat: torch.Tensor,
    qubits: list[int] | tuple[int],
    n_qubits: int,
    batch_size: int,
) -> torch.Tensor:
    """
    Vmap a batched gate over a batched state and
    apply the matrix 'mat' to `state`.

    Arguments:
        state (torch.Tensor): input state of shape [2] * N_qubits + [batch_size]
        mat (torch.Tensor): the matrix representing the gate
        qubits (list, tuple, array): iterator containing the qubits
        the gate is applied to
        n_qubits (int): the total number of qubits of the system
        batch

    Returns:
     state (torch.Tensor): the quantum state after application of the gate.
            Same shape as `ìnput_state`
    """

    def _apply(
        state: torch.Tensor,
        mat: torch.Tensor,
        qubits: list[int] | tuple[int],
        n_qubits: int,
    ) -> torch.Tensor:
        mat = mat.view([2] * len(qubits) * 2)
        mat_dims = list(range(len(qubits), 2 * len(qubits)))
        state_dims = list(qubits)
        axes = (mat_dims, state_dims)
        state = torch.tensordot(mat, state, dims=axes)
        inv_perm = torch.argsort(
            torch.tensor(
                state_dims + [j for j in range(n_qubits) if j not in state_dims], dtype=torch.int
            )
        )
        state = torch.permute(state, tuple(inv_perm))
        return state

    return torch.vmap(
        lambda s, m: _apply(state=s, mat=m, qubits=qubits, n_qubits=n_qubits),
        in_dims=(len(state.size()) - 1, len(mat.size()) - 1),
        out_dims=len(state.size()) - 1,
    )(state, mat)



"""Test examples scripts."""
from __future__ import annotations

import os
import subprocess
import sys
from pathlib import Path
from typing import List

import pytest

expected_fail = {}  # type: ignore
skip = {}  # type: ignore


def get_ipynb_files(dir: Path) -> List[Path]:
    files = []

    for it in dir.iterdir():
        if it.suffix == ".ipynb":
            files.append(it)
        elif it.is_dir() and it.name != "deprecated":
            files.extend(get_ipynb_files(it))
    return files


# FIXME: refactor choice of notebooks folders
notebooks_dir = Path(__file__).parent.parent.joinpath("docs").resolve()
assert notebooks_dir.exists()
notebooks = get_ipynb_files(notebooks_dir)
notebooks_names = [f"{example.relative_to(notebooks_dir)}" for example in notebooks]
for example, reason in expected_fail.items():
    try:
        notebooks[notebooks_names.index(example)] = pytest.param(  # type: ignore
            example, marks=pytest.mark.xfail(reason=reason)
        )
    except ValueError:
        pass

for example, reason in skip.items():
    try:
        notebooks[notebooks_names.index(example)] = pytest.param(  # type: ignore
            example, marks=pytest.mark.skip(reason=reason)
        )
    except ValueError:
        pass


@pytest.mark.parametrize("notebook", notebooks, ids=notebooks_names)
def test_notebooks(notebook: Path) -> None:
    """Execute docs notebooks as a test, passes if it returns 0."""
    jupyter_cmd = ["-m", "jupyter", "nbconvert", "--to", "python", "--execute"]
    cmd = [sys.executable, *jupyter_cmd, notebook]
    with subprocess.Popen(
        cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, env={**os.environ}  # type: ignore
    ) as run_example:
        stdout, stderr = run_example.communicate()
        error_string = (
            f"Notebook {notebook.name} failed\n"
            f"stdout:{stdout.decode()}\n"
            f"stderr: {stderr.decode()}"
        )

    if run_example.returncode != 0:
        raise Exception(error_string)



# import time
# import pyqtorch.modules as pyq
# from pyqtorch.modules.utils import _apply_parallel
# # from pyqtorch.core.utils import _apply_parallel
# import torch



# N_QUBITS = 16
# def hea(n_qubits:int, depth:int) -> list:
#     for d in range(depth):
#         rxgates = [pyq.RX([i], N_QUBITS) for i in range(N_QUBITS)]
#         rygates = [pyq.RY([i], N_QUBITS) for i in range(N_QUBITS)]
#         rzgates = [pyq.RZ([i], N_QUBITS) for i in range(N_QUBITS)]

# circ = pyq.QuantumCircuit(N_QUBITS, rxgates+rygates+rzgates)

# thetas = torch.rand(1)
# state = pyq.zero_state(N_QUBITS)


# def circ_fwd(state) -> torch.Tensor:
#     return circ(state, thetas)


# def parallel_(state) -> torch.Tensor:
#     state = _apply_parallel(state, thetas, rxgates, N_QUBITS)
#     state = _apply_parallel(state, thetas, rygates, N_QUBITS)
#     state = _apply_parallel(state, thetas, rzgates, N_QUBITS)
#     return state


# def timeit(func):
#     def wrapper(*args, **kwargs):
#         start_time = time.time()
#         result = func(*args, **kwargs)
#         end_time = time.time()
#         print(f"{func.__name__} took {end_time - start_time} seconds to run.")
#         return result
#     return wrapper


# @timeit
# def time_fn():
#     circ_fwd(state)


# if __name__ == "__main__":
#     time_fn()



from __future__ import annotations

from pyqtorch.modules.circuit import (
    EntanglingLayer,
    FeaturemapLayer,
    QuantumCircuit,
    VariationalLayer,
)
from pyqtorch.modules.hamevo import HamEvo, HamEvoEig, HamEvoExp, HamEvoType, HamiltonianEvolution
from pyqtorch.modules.parametric import CPHASE, CRX, CRY, CRZ, PHASE, RX, RY, RZ, U
from pyqtorch.modules.primitive import (
    CNOT,
    CSWAP,
    CY,
    CZ,
    SWAP,
    H,
    I,
    N,
    S,
    SDagger,
    T,
    Toffoli,
    X,
    Y,
    Z,
)
from pyqtorch.modules.utils import (
    _apply_parallel,
    flatten_wf,
    invert_endianness,
    is_normalized,
    normalize,
    overlap,
    random_state,
    uniform_state,
    zero_state,
)



# Copyright 2022 PyQ Development Team

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from string import ascii_letters as ABC

import numpy as np
import torch
from numpy.typing import NDArray

ABC_ARRAY: NDArray = np.array(list(ABC))


def _apply_gate(
    state: torch.Tensor,
    mat: torch.Tensor,
    qubits: list[int] | tuple[int],
    N_qubits: int,
) -> torch.Tensor:
    """
    Apply the matrix 'mat' corresponding to a gate to `state`.

    Arguments:
        state (torch.Tensor): input state of shape [2] * N_qubits + [batch_size]
        mat (torch.Tensor): the matrix representing the gate
        qubits (list, tuple, array): iterator containing the qubits
        the gate is applied to
        N_qubits (int): the total number of qubits of the system

    Returns:
     state (torch.Tensor): the quantum state after application of the gate.
            Same shape as `ìnput_state`
    """
    mat = mat.view([2] * len(qubits) * 2)
    mat_dims = list(range(len(qubits), 2 * len(qubits)))
    qubits = list(qubits)
    axes = (mat_dims, qubits)
    state = torch.tensordot(mat, state, dims=axes)
    inv_perm = torch.argsort(
        torch.tensor(qubits + [j for j in range(N_qubits + 1) if j not in qubits], dtype=torch.int)
    )
    state = torch.permute(state, tuple(inv_perm))
    return state


def _apply_einsum_gate(
    state: torch.Tensor, mat: torch.Tensor, qubits: list[int] | tuple[int], N_qubits: int
) -> torch.Tensor:
    """
    Same as `apply_gate` but with the `torch.einsum` function

    Arguments:
        state (torch.Tensor): input state of shape [2] * N_qubits + [batch_size]
        mat (torch.Tensor): the matrix representing the gate
        qubits (list, tuple, array): Sized iterator containing the qubits
                                     the gate is applied to
        N_qubits (int): the total number of qubits of the system

    Returns:
        state (torch.Tensor): The state after application of the gate.
                              Same shape as `ìnput_state`
    """
    mat = mat.reshape([2] * len(qubits) * 2)
    state_indices = ABC_ARRAY[0 : N_qubits + 1]
    qubits = list(qubits)
    # Create new indices for the matrix indices
    mat_indices = ABC_ARRAY[N_qubits + 2 : N_qubits + 2 + 2 * len(qubits)]
    mat_indices[len(qubits) : 2 * len(qubits)] = state_indices[qubits]

    # Create the new state indices: same as input states but
    # modified affected qubits
    new_state_indices = state_indices.copy()
    new_state_indices[qubits] = mat_indices[0 : len(qubits)]

    # Transform the arrays into strings
    state_indices = "".join(list(state_indices))
    new_state_indices = "".join(list(new_state_indices))
    mat_indices = "".join(list(mat_indices))

    einsum_indices = f"{mat_indices},{state_indices}->{new_state_indices}"

    state = torch.einsum(einsum_indices, mat, state)

    return state


def _apply_batch_gate(
    state: torch.Tensor,
    mat: torch.Tensor,
    qubits: list[int] | tuple[int],
    N_qubits: int,
    batch_size: int,
) -> torch.Tensor:
    """
    Apply a batch execution of gates to a circuit.
    Given an tensor of states [state_0, ... state_b] and
    an tensor of gates [G_0, ... G_b] it will return the
    tensor [G_0 * state_0, ... G_b * state_b]. All gates
    are applied to the same qubit.

    Arguments:
        state (torch.Tensor): input state of shape [2] * N_qubits + [batch_size]
        mat (torch.Tensor): the tensor representing the gates.
            The last dimension is the batch dimension.
            It has to be the sam eas the last dimension of `state`
        qubits (list, tuple, array): Sized iterator containing the qubits
                                        the gate is applied to.
        N_qubits (int): the total number of qubits of the system
        batch_size (int): the size of the batch

    Returns:
        torch.Tensor: The state after application of the gate.
                                Same shape as `input_state`.
    Examples:
    ```python exec="on" source="above" result="json"
    import torch
    import pyqtorch.modules as pyq

    state = pyq.zero_state(n_qubits=2)
    print(state)  #tensor([[[1.+0.j],[0.+0.j]],[[0.+0.j],[0.+0.j]]], dtype=torch.complex128)
    ```
    """
    mat = mat.view([2] * len(qubits) * 2 + [batch_size])

    state_indices = ABC_ARRAY[0 : N_qubits + 1].copy()
    qubits = list(qubits)
    mat_indices = ABC_ARRAY[N_qubits + 2 : N_qubits + 2 + 2 * len(qubits) + 1].copy()
    mat_indices[len(qubits) : 2 * len(qubits)] = state_indices[qubits]
    mat_indices[-1] = state_indices[-1]

    new_state_indices = state_indices.copy()
    new_state_indices[qubits] = mat_indices[0 : len(qubits)]

    state_indices = "".join(list(state_indices))
    new_state_indices = "".join(list(new_state_indices))
    mat_indices = "".join(list(mat_indices))

    einsum_indices = f"{mat_indices},{state_indices}->{new_state_indices}"

    state = torch.einsum(einsum_indices, mat, state)

    return state



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any

import torch
from numpy.typing import ArrayLike
from torch.nn import Module

import pyqtorch.modules as pyq


class AbstractGate(ABC, Module):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        super().__init__()
        self.qubits = qubits
        self.n_qubits = n_qubits

    def __mul__(self, other: AbstractGate | pyq.QuantumCircuit) -> pyq.QuantumCircuit:
        if isinstance(other, AbstractGate):
            ml = torch.nn.ModuleList([self, other])
            return pyq.QuantumCircuit(max(self.n_qubits, other.n_qubits), ml)
        elif isinstance(other, pyq.QuantumCircuit):
            ml = torch.nn.ModuleList([self]) + other.operations
            return pyq.QuantumCircuit(max(self.n_qubits, other.n_qubits), ml)
        else:
            return TypeError(f"Cannot compose {type(self)} with {type(other)}")

    def __key(self) -> tuple:
        return (self.n_qubits, *self.qubits)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, type(self)):
            return self.__key() == other.__key()
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__key())

    @abstractmethod
    def matrices(self, tensors: torch.Tensor) -> torch.Tensor:
        # NOTE: thetas are assumed to be of shape (1,batch_size) or (batch_size,) because we
        # want to allow e.g. (3,batch_size) in the U gate.
        ...

    @abstractmethod
    def apply(self, matrix: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        ...

    @abstractmethod
    def forward(self, state: torch.Tensor, thetas: torch.Tensor) -> torch.Tensor:
        ...

    def extra_repr(self) -> str:
        return f"qubits={self.qubits}, n_qubits={self.n_qubits}"



from __future__ import annotations

import math
from typing import Callable

import pytest
import torch
from torch import Tensor

import pyqtorch.core as func_pyq
import pyqtorch.modules as pyq
from pyqtorch.matrices import OPERATIONS_DICT
from pyqtorch.modules.abstract import AbstractGate

DEVICE = "cuda" if torch.cuda.is_available() else "cpu"
DTYPE = torch.cdouble

state_000 = pyq.zero_state(3, device=DEVICE, dtype=DTYPE)
state_001 = pyq.X(qubits=[2], n_qubits=3)(state_000)
state_100 = pyq.X(qubits=[0], n_qubits=3)(state_000)
state_101 = pyq.X(qubits=[2], n_qubits=3)(pyq.X(qubits=[0], n_qubits=3)(state_000))
state_110 = pyq.X(qubits=[1], n_qubits=3)(pyq.X(qubits=[0], n_qubits=3)(state_000))
state_111 = pyq.X(qubits=[2], n_qubits=3)(
    pyq.X(qubits=[1], n_qubits=3)(pyq.X(qubits=[0], n_qubits=3)(state_000))
)

state_0000 = pyq.zero_state(4, device=DEVICE, dtype=DTYPE)
state_1110 = pyq.X(qubits=[0], n_qubits=4)(
    pyq.X(qubits=[1], n_qubits=4)(pyq.X(qubits=[2], n_qubits=4)(state_0000))
)
state_1111 = pyq.X(qubits=[0], n_qubits=4)(
    pyq.X(qubits=[1], n_qubits=4)(
        pyq.X(qubits=[2], n_qubits=4)(pyq.X(qubits=[3], n_qubits=4)(state_0000))
    )
)


@pytest.mark.parametrize("batch_size", [i for i in range(1, 2, 10)])
@pytest.mark.parametrize("n_qubits", [i for i in range(1, 6)])
@pytest.mark.parametrize("gate", ["X", "Y", "Z", "H", "I", "N", "S", "T", "SDagger"])
def test_constant_gates(batch_size: int, n_qubits: int, gate: str) -> None:
    dtype = torch.cdouble
    qubits = [torch.randint(low=0, high=n_qubits, size=(1,)).item()]

    state = pyq.random_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    Op = getattr(pyq, gate)
    FuncOp = getattr(func_pyq.operation, gate)

    func_out = FuncOp(state, qubits, n_qubits)
    op = Op(qubits, n_qubits).to(device=DEVICE, dtype=dtype)
    mod_out = op(state)

    assert torch.allclose(func_out, mod_out)


@pytest.mark.parametrize("batch_size", [i for i in range(1, 2, 10)])
@pytest.mark.parametrize("n_qubits", [i for i in range(1, 6)])
@pytest.mark.parametrize("gate", ["RX", "RY", "RZ"])
def test_parametrized_gates(batch_size: int, n_qubits: int, gate: str) -> None:
    qubits = [torch.randint(low=0, high=n_qubits, size=(1,)).item()]

    state = pyq.random_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    phi = torch.rand(batch_size, device=DEVICE, dtype=DTYPE)

    Op = getattr(pyq, gate)
    FuncOp = getattr(func_pyq.batched_operation, f"batched{gate}")

    func_out = FuncOp(phi, state, qubits, n_qubits)
    op = Op(qubits, n_qubits).to(device=DEVICE, dtype=DTYPE)
    mod_out = op(state, phi)

    assert torch.allclose(func_out, mod_out)


@pytest.mark.parametrize("batch_size", [i for i in range(1, 2, 10)])
@pytest.mark.parametrize("n_qubits", [i for i in range(2, 6)])
@pytest.mark.parametrize("gate", ["CRX", "CRY", "CRZ"])
def test_controlled_parametrized_gates(batch_size: int, n_qubits: int, gate: str) -> None:
    qubits = torch.randint(low=0, high=n_qubits, size=(2,))

    while qubits[0] == qubits[1]:
        qubits[1] = torch.randint(low=0, high=n_qubits, size=(1,))

    qubits = [qubits[0].item(), qubits[1].item()]

    state = pyq.random_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    phi = torch.rand(batch_size, device=DEVICE, dtype=DTYPE)

    Op = getattr(pyq, gate)
    BatchedOP = getattr(func_pyq.batched_operation, f"batched{gate}")

    func_out = BatchedOP(phi, state, qubits, n_qubits)
    op = Op(qubits, n_qubits).to(device=DEVICE, dtype=DTYPE)
    mod_out = op(state, phi)

    assert torch.allclose(func_out, mod_out)


@pytest.mark.parametrize("batch_size", [i for i in range(1, 2, 10)])
@pytest.mark.parametrize("n_qubits", [i for i in range(2, 6)])
def test_circuit(batch_size: int, n_qubits: int) -> None:
    ops = [
        pyq.X([0], n_qubits),
        pyq.X([1], n_qubits),
        pyq.RX([1], n_qubits),
        pyq.CNOT([0, 1], n_qubits),
    ]
    circ = pyq.QuantumCircuit(n_qubits, ops).to(device=DEVICE, dtype=DTYPE)

    state = pyq.random_state(n_qubits, batch_size)
    phi = torch.rand(batch_size, device=DEVICE, dtype=DTYPE, requires_grad=True)

    assert circ(state, phi).size() == tuple(2 for _ in range(n_qubits)) + (batch_size,)

    state = pyq.random_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)

    res = circ(state, phi)
    assert not torch.all(torch.isnan(res))
    dres_theta = torch.autograd.grad(res, phi, torch.ones_like(res), create_graph=True)[0]
    assert not torch.all(torch.isnan(dres_theta))


@pytest.mark.parametrize("batch_size", [1, 2, 4, 6])
def test_empty_circuit(batch_size: int) -> None:
    n_qubits = 2
    ops: list = []
    circ = pyq.QuantumCircuit(n_qubits, ops).to(device=DEVICE, dtype=DTYPE)

    state = circ.init_state(batch_size)
    phi = torch.rand(batch_size, device=DEVICE, dtype=DTYPE, requires_grad=True)

    assert circ(state, phi).size() == (2, 2, batch_size)

    state = pyq.random_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)

    res = circ(state, phi)
    assert not torch.all(torch.isnan(res))


@pytest.mark.parametrize("batch_size", [1, 2, 4, 6])
def test_U_gate(batch_size: int) -> None:
    n_qubits = 1
    u = pyq.U([0], n_qubits)
    x = torch.rand(3, batch_size)
    state = pyq.random_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    assert not torch.all(torch.isnan(u(state, x)))


@pytest.mark.parametrize("batch_size", [1, 2, 4, 6])
def test_N_gate(batch_size: int) -> None:
    n_qubits = 1
    n = pyq.N([0], n_qubits)
    init_state = pyq.zero_state(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    final_state = pyq.X([0], n_qubits)(init_state)
    assert torch.equal(final_state, n(final_state))


@pytest.mark.parametrize(
    "a,b,val",
    [
        (pyq.X([0], 1), pyq.X([0], 1), True),
        (pyq.RY([0], 1), pyq.RY([0], 1), True),
        (pyq.RY([0], 1), pyq.RY([0], 1), True),
        (pyq.X([0], 1), pyq.Y([0], 1), False),
        (pyq.RX([0], 2), pyq.RX([0], 1), False),
        (pyq.RX([1], 1), pyq.RX([0], 1), False),
    ],
)
def test_gate_equality(a: AbstractGate, b: AbstractGate, val: bool) -> None:
    x = a == b
    assert x == val


def test_circuit_equality() -> None:
    c1 = pyq.QuantumCircuit(2, [pyq.RX([0], 2), pyq.Z([1], 2)])
    c2 = pyq.QuantumCircuit(2, [pyq.RX([0], 2), pyq.Z([1], 2)])
    assert c1 == c2

    c3 = pyq.QuantumCircuit(2, [pyq.RX([1], 2), pyq.Z([1], 2)])
    assert c1 != c3

    c4 = pyq.QuantumCircuit(2, [pyq.Z([1], 2)])
    assert c3 != c4

    c5 = pyq.QuantumCircuit(2, [pyq.RX([0], 2)])
    assert c1 != c5


@pytest.mark.parametrize("n_qubits", [1, 2, 3])
def test_gate_composition(n_qubits: int) -> None:
    x = pyq.X(torch.randint(0, n_qubits, (1,)).tolist(), n_qubits)
    y = pyq.Y(torch.randint(0, n_qubits, (1,)).tolist(), n_qubits)
    circ = x * y
    truth = pyq.QuantumCircuit(n_qubits, [x, y])
    assert circ == truth

    rx = pyq.RX(torch.randint(0, n_qubits, (1,)).tolist(), n_qubits)
    ry = pyq.RY(torch.randint(0, n_qubits, (1,)).tolist(), n_qubits)
    circ = rx * ry
    truth = pyq.QuantumCircuit(n_qubits, [rx, ry])
    assert circ == truth

    r1, r2, r3 = 1, n_qubits, max(n_qubits - 1, 1)
    z1 = pyq.Z(torch.randint(0, r1, (1,)).tolist(), r1)
    y = pyq.Y(torch.randint(0, r2, (1,)).tolist(), r2)
    z2 = pyq.Z(torch.randint(0, r3, (1,)).tolist(), r3)
    circ = z1 * y * z2
    truth = pyq.QuantumCircuit(max(r1, r2, r3), [z1, y, z2])
    assert circ == truth

    rr1, rr2, rr3 = 1, n_qubits, max(n_qubits - 1, 1)
    rz1 = pyq.RZ(torch.randint(0, rr1, (1,)).tolist(), rr1)
    ry1 = pyq.RY(torch.randint(0, rr2, (1,)).tolist(), rr2)
    rz2 = pyq.RZ(torch.randint(0, rr3, (1,)).tolist(), rr3)

    circ = rz1 * ry1 * rz2
    truth = pyq.QuantumCircuit(max(r1, r2, r3), [rz1, ry1, rz2])
    assert circ == truth


@pytest.mark.parametrize("n_qubits", [1, 2, 3])
def test_circuit_composition(n_qubits: int) -> None:
    r = torch.randint(1, n_qubits + 1, (1,)).tolist()
    rx = pyq.RX(r, n_qubits)

    circ = rx * pyq.QuantumCircuit(1, [pyq.X([0], 1), pyq.Y([0], 1)])
    truth = pyq.QuantumCircuit(n_qubits, [rx, pyq.X([0], 1), pyq.Y([0], 1)])
    assert circ == truth

    circ = pyq.QuantumCircuit(1, [pyq.X([0], 1), pyq.Y([0], 1)]) * rx
    truth = pyq.QuantumCircuit(n_qubits, [pyq.X([0], 1), pyq.Y([0], 1), rx])
    assert circ == truth

    circ = pyq.QuantumCircuit(1, [pyq.X([0], 1), pyq.Y([0], 1)]) * pyq.QuantumCircuit(
        n_qubits, [rx, pyq.RY([0], 1)]
    )
    truth = pyq.QuantumCircuit(n_qubits, [pyq.X([0], 1), pyq.Y([0], 1), rx, pyq.RY([0], 1)])
    assert circ == truth


@pytest.mark.parametrize(
    "initial_state,expected_state",
    [
        (state_000, state_000),
        (state_001, state_001),
        (state_100, state_100),
        (state_101, state_110),
        (state_110, state_101),
    ],
)
def test_CSWAP_controlqubits0(initial_state: Tensor, expected_state: Tensor) -> None:
    n_qubits = 3
    cswap = pyq.CSWAP([0, 1, 2], n_qubits)
    assert torch.allclose(cswap(initial_state), expected_state)


@pytest.mark.parametrize(
    "initial_state,expected_state",
    [
        (state_000, state_000),
        (state_001, state_001),
        (state_100, state_100),
        (state_101, state_101),
        (state_110, state_111),
        (state_1110, state_1111),
    ],
)
def test_Toffoli_controlqubits0(initial_state: Tensor, expected_state: Tensor) -> None:
    n_qubits = int(math.log2(torch.numel(initial_state)))
    toffoli = pyq.Toffoli(range(n_qubits), n_qubits)
    assert torch.allclose(toffoli(initial_state), expected_state)


@pytest.mark.parametrize(
    "initial_state,expects_rotation",
    [
        (state_000, False),
        (state_001, False),
        (state_100, False),
        (state_101, False),
        (state_110, True),
        (state_1110, True),
    ],
)
@pytest.mark.parametrize("gate", ["RX", "RY", "RZ", "PHASE"])
@pytest.mark.parametrize("batch_size", [i for i in range(1, 2, 10)])
def test_multi_controlled_gates(
    initial_state: Tensor, expects_rotation: bool, batch_size: int, gate: str
) -> None:
    rot_gate = getattr(pyq, gate)
    controlled_rot_gate = getattr(pyq, "C" + gate)
    phi = torch.rand(batch_size, device=DEVICE, dtype=DTYPE)
    n_qubits = int(math.log2(torch.numel(initial_state)))
    op = controlled_rot_gate(range(n_qubits), n_qubits).to(device=DEVICE, dtype=DTYPE)
    out = op(initial_state, phi)
    expected_state = (
        rot_gate([n_qubits - 1], n_qubits)(initial_state, phi)
        if expects_rotation
        else initial_state
    )
    assert torch.allclose(out, expected_state)


@pytest.mark.parametrize("state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state])
@pytest.mark.parametrize("n_qubits", [i for i in range(1, 8)])
@pytest.mark.parametrize("batch_size", [i for i in range(1, 8)])
def test_isnormalized_states(state_fn: Callable, n_qubits: int, batch_size: int) -> None:
    state = state_fn(n_qubits, batch_size, device=DEVICE, dtype=DTYPE)
    assert pyq.is_normalized(state)


@pytest.mark.parametrize("n_qubits", [i for i in range(1, 8)])
@pytest.mark.parametrize("batch_size", [i for i in range(1, 8)])
def test_state_shapes(n_qubits: int, batch_size: int) -> None:
    zero = pyq.zero_state(n_qubits, batch_size, device=DEVICE, dtype=DTYPE)
    uni = pyq.uniform_state(n_qubits, batch_size, device=DEVICE, dtype=DTYPE)
    rand = pyq.random_state(n_qubits, batch_size, device=DEVICE, dtype=DTYPE)
    assert zero.shape == rand.shape and uni.shape == rand.shape


@pytest.mark.parametrize("state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state])
@pytest.mark.parametrize("n_qubits", [i for i in range(1, 8)])
@pytest.mark.parametrize("batch_size", [i for i in range(1, 8)])
def test_overlap_states_batch_nqubits(state_fn: Callable, n_qubits: int, batch_size: int) -> None:
    state = state_fn(n_qubits, batch_size, device=DEVICE, dtype=DTYPE)
    assert torch.allclose(
        pyq.overlap(state, state),
        torch.ones(batch_size),
    )


@pytest.mark.parametrize("state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state])
@pytest.mark.parametrize("batch_size", [i for i in range(1, 2, 10)])
@pytest.mark.parametrize("n_qubits", [i for i in range(1, 6)])
def test_parametrized_phase_gate(state_fn: Callable, batch_size: int, n_qubits: int) -> None:
    qubits = [torch.randint(low=0, high=n_qubits, size=(1,)).item()]
    state = state_fn(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    phi = torch.tensor([torch.pi / 2], dtype=torch.cdouble)
    phase = pyq.PHASE(qubits, n_qubits).to(device=DEVICE, dtype=DTYPE)
    constant_phase = pyq.S(qubits, n_qubits).to(device=DEVICE, dtype=DTYPE)
    assert torch.allclose(phase(state, phi), constant_phase(state, phi))


@pytest.mark.parametrize("state_fn", [pyq.random_state, pyq.zero_state, pyq.uniform_state])
def test_parametric_phase_hamevo(
    state_fn: Callable, batch_size: int = 1, n_qubits: int = 1
) -> None:
    qubits = [0]
    state = state_fn(n_qubits, batch_size=batch_size, device=DEVICE, dtype=DTYPE)
    phi = torch.rand(1, dtype=torch.cdouble)
    H = (OPERATIONS_DICT["Z"] - OPERATIONS_DICT["I"]) / 2
    hamevo = pyq.HamEvoExp(H, phi, qubits=qubits, n_qubits=n_qubits)
    phase = pyq.PHASE(qubits, n_qubits).to(device=DEVICE, dtype=DTYPE)
    assert torch.allclose(phase(state, phi), hamevo(state))



from __future__ import annotations

from typing import Any

import torch
from torch.nn import Module, ModuleList, Parameter, init

from pyqtorch.modules.abstract import AbstractGate
from pyqtorch.modules.primitive import CNOT
from pyqtorch.modules.utils import zero_state


class QuantumCircuit(Module):
    def __init__(self, n_qubits: int, operations: list):
        """
        Creates a QuantumCircuit that can be used to compose multiple gates
        from a list of operations.

        Arguments:
            n_qubits (int): The total number of qubits in the circuit.
            operations (list): A list of gate operations to be applied in the circuit.

        Example:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        #create a circuit with 2 qubits than provide a list of operations .
        #in this example we apply a X gate followed by a CNOT gate.
        circ = pyq.QuantumCircuit(
                                    n_qubits=2,
                                    operations=[
                                        pyq.X([0], 2),
                                        pyq.CNOT([0,1], 2)
                                    ]
                                )
        #create a zero state
        z = pyq.zero_state(2)

        #apply the circuit and its list of operations onto the zero state
        result=circ(z)

        #print the result
        print(result) #tensor([[[0.+0.j],[0.+0.j]],[[0.+0.j],[1.+0.j]]], dtype=torch.complex128)
        ```
        """
        super().__init__()
        self.n_qubits = n_qubits
        self.operations = torch.nn.ModuleList(operations)

    def __mul__(self, other: AbstractGate | QuantumCircuit) -> QuantumCircuit:
        if isinstance(other, QuantumCircuit):
            n_qubits = max(self.n_qubits, other.n_qubits)
            return QuantumCircuit(n_qubits, self.operations.extend(other.operations))

        if isinstance(other, AbstractGate):
            n_qubits = max(self.n_qubits, other.n_qubits)
            return QuantumCircuit(n_qubits, self.operations.append(other))

        else:
            return ValueError(f"Cannot compose {type(self)} with {type(other)}")

    def __key(self) -> tuple:
        return (self.n_qubits, *self.operations)

    def __eq__(self, other: Any) -> bool:
        if isinstance(other, QuantumCircuit):
            return self.__key() == other.__key()
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__key())

    def forward(self, state: torch.Tensor, thetas: torch.Tensor = None) -> torch.Tensor:
        """
        Forward pass of the quantum circuit.

        Arguments:
            state (torch.Tensor): The input quantum state tensor.
            thetas (torch.Tensor): Optional tensor of parameters for the circuit operations.

        Returns:
            torch.Tensor: The output quantum state tensor after applying the circuit operations.

        """
        for op in self.operations:
            state = op(state, thetas)
        return state

    @property
    def _device(self) -> torch.device:
        try:
            (_, buffer) = next(self.named_buffers())
            return buffer.device
        except StopIteration:
            return torch.device("cpu")

    def init_state(self, batch_size: int) -> torch.Tensor:
        return zero_state(self.n_qubits, batch_size, device=self._device)


def FeaturemapLayer(n_qubits: int, Op: Any) -> QuantumCircuit:
    """
    Creates a feature map layer in a quantum neural network.
    The FeaturemapLayer is a convenience constructor for a QuantumCircuit
    which accepts an operation to put on every qubit.

    Arguments:
        n_qubits (int): The total number of qubits in the circuit.
        Op (Any): The quantum operation to be applied in the feature map layer.

    Returns:
        QuantumCircuit: The feature map layer represented as a QuantumCircuit.

    Example:
    ```python exec="on" source="above" result="json"
    import torch
    import pyqtorch.modules as pyq

    #create a FeaturemapLayer to apply the RX operation on all 3 Qubits
    circ = pyq.FeaturemapLayer(n_qubits=3, Op=pyq.RX)
    print(circ)

    states = pyq.zero_state(n_qubits=3, batch_size=4)
    inputs = torch.rand(4)

    # the same batch of inputs are passed to the operations
    circ(states, inputs).shape
    ```
    """
    operations = [Op([i], n_qubits) for i in range(n_qubits)]
    return QuantumCircuit(n_qubits, operations)


class VariationalLayer(QuantumCircuit):
    def __init__(self, n_qubits: int, Op: Any):
        """
        Represents a variational layer in a quantum neural network allowing you
        to create a trainable QuantumCircuit.
        If you want the angles of your circuit to be trainable you can use a VariationalLayer.
        The VariationalLayer ignores the second input (because it has trainable angle parameters).

        Arguments:
            n_qubits (int): The total number of qubits in the circuit.
            Op (Any): The quantum operation to be applied in the variational layer.


        Example:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq
        #create a variational layer with 3 qubits and operation of RX as the second parameter
        circ = pyq.VariationalLayer(n_qubits=3, Op=pyq.RX)
        state = pyq.zero_state(3)
        this_argument_is_ignored = None
        result=circ(state, this_argument_is_ignored)
        print(result)
        ```
        """
        operations = ModuleList([Op([i], n_qubits) for i in range(n_qubits)])
        super().__init__(n_qubits, operations)

        self.thetas = Parameter(torch.empty(n_qubits, Op.n_params))
        self.reset_parameters()

    def reset_parameters(self) -> None:
        init.uniform_(self.thetas, -2 * torch.pi, 2 * torch.pi)

    def forward(self, state: torch.Tensor, _: torch.Tensor = None) -> torch.Tensor:
        for op, t in zip(self.operations, self.thetas):
            state = op(state, t)
        return state


class EntanglingLayer(QuantumCircuit):
    def __init__(self, n_qubits: int):
        """
        Represents an entangling layer in a quantum neural network by entangling Qubits

        Args:
            n_qubits (int): The total number of qubits in the circuit.

        Example:
        ```python exec="on" source="above" result="json"
        from pyqtorch.modules.circuit import EntanglingLayer

        # Create an entangling layer with 4 qubits
        entangling_layer = EntanglingLayer(n_qubits=4)

        print(entangling_layer)
        ```
        """
        operations = ModuleList(
            [CNOT([i % n_qubits, (i + 1) % n_qubits], n_qubits) for i in range(n_qubits)]
        )
        super().__init__(n_qubits, operations)



from __future__ import annotations

import math

import torch
from numpy.typing import ArrayLike

from pyqtorch.core.operation import _apply_gate, create_controlled_matrix_from_operation
from pyqtorch.matrices import OPERATIONS_DICT
from pyqtorch.modules.abstract import AbstractGate


class PrimitiveGate(AbstractGate):
    def __init__(self, gate: str, qubits: ArrayLike, n_qubits: int):
        super().__init__(qubits, n_qubits)
        self.gate = gate
        self.register_buffer("matrix", OPERATIONS_DICT[gate])

    def matrices(self, _: torch.Tensor) -> torch.Tensor:
        return self.matrix

    def apply(self, matrix: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        return _apply_gate(state, matrix, self.qubits, self.n_qubits)

    def forward(self, state: torch.Tensor, _: torch.Tensor = None) -> torch.Tensor:
        return self.apply(self.matrix, state)


class X(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents an X gate (Pauli-X gate) in a quantum circuit.
        The X gate class creates a  X gate that performs a PI rotation around the X axis

        Arguments:
            qubits (ArrayLike): The qubits to which the X gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an X gate
        x_gate = pyq.X(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the X gate to the zero state
        result = x_gate(z_state)
        print(result)
        ```
        """
        super().__init__("X", qubits, n_qubits)


class Y(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a Y gate (Pauli-Y gate) in a quantum circuit.
        The Y gate class creates a  Y gate that performs a PI rotation around the Y axis.


        Arguments:
            qubits (ArrayLike): The qubits to which the Y gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a Y gate
        y_gate = pyq.Y(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the Y gate to the zero state
        result = y_gate(z_state)
        print(result)
        ```
        """
        super().__init__("Y", qubits, n_qubits)


class Z(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a Z gate (Pauli-Z gate) in a quantum circuit.
        The ZGate class creates a Z gate that performs a PI rotation around the Z axis.

        Arguments:
            qubits (ArrayLike): The qubits to which the Z gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a Z gate
        z_gate = pyq.Z(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the Z gate to the zero state
        result = z_gate(z_state)

        print(result)
        ```
        """
        super().__init__("Z", qubits, n_qubits)


# FIXME: do we really have to apply a matrix here?
# can't we just return the identical state?
class I(PrimitiveGate):  # noqa: E742
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents an I gate (identity gate) in a quantum circuit.
        The I gate class creates a I gate, which has no effect on the state of a qubit.

        Arguments:
            qubits (ArrayLike): The qubits to which the I gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an I gate
        i_gate = pyq.I(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the I gate to the zero state
        result = i_gate(z_state)
        print(result)
        ```
        """
        super().__init__("I", qubits, n_qubits)


class H(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents an H gate (Hadamard gate) in a quantum circuit.
        The H Gate class creates a H gate. It performs a PI rotation
        around the X+Z axis changing the basis from |0⟩,|1⟩ to  |+⟩,|-⟩
        and from |+⟩,|-⟩ back to |0⟩,|1⟩  depending on the number of times
        the gate is applied

        Arguments:
            qubits (ArrayLike): The qubits to which the H gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an H gate
        h_gate = pyq.H(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the H gate to the zero state
        result = h_gate(z_state)
        print(result)  # tensor([[0.7071+0.j],[0.7071+0.j]], dtype=torch.complex128)
        ```
        """
        super().__init__("H", qubits, n_qubits)


class T(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a T gate (PI/4 phase gate) in a quantum circuit.

        Arguments:
            qubits (ArrayLike): The qubits to which the T gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a T gate
        t_gate = pyq.T(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the T gate to the zero state
        result = t_gate(z_state)
        print(result)
        ```
        """
        super().__init__("T", qubits, n_qubits)


class S(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents an S gate (PI/2 phase gate) in a quantum circuit.

        Arguments:
            qubits (ArrayLike): The qubits to which the S gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an S gate
        s_gate = pyq.S(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the S gate to the zero state
        result = s_gate(z_state)
        print(result)
        ```
        """
        super().__init__("S", qubits, n_qubits)


class SDagger(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents an SDagger gate (-PI/2 phase gate) in a quantum circuit.

        Arguments:
            qubits (ArrayLike): The qubits to which the S gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an SDagger gate
        sdagger_gate = pyq.SDagger(qubits=[0], n_qubits=1)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=1)

        # Apply the SDagger gate to the zero state
        result = sdagger_gate(z_state)
        print(result)
        ```
        """
        super().__init__("SDAGGER", qubits, n_qubits)


class N(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents an N gate ((I-Z)/2 projector gate) in a quantum circuit.

        Arguments:
            qubits (ArrayLike): The qubits to which the S gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create an N gate
        n_gate = pyq.N(qubits=[0], n_qubits=1)

        # Create a zero state
        n_state = pyq.zero_state(n_qubits=1)

        # Apply the N gate to the zero state
        result = n_gate(n_state)
        print(result)
        ```
        """
        super().__init__("N", qubits, n_qubits)


class SWAP(PrimitiveGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a SWAP gate in a quantum circuit.
        The SwapGate swaps the qubit states of two quantum wires.


        Arguments:
            qubits (ArrayLike): The qubits to which the SWAP gate is applied.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a SWAP gate
        swap_gate = pyq.SWAP(qubits=[0, 1], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        # Apply the SWAP gate to the zero state
        result = swap_gate(z_state)
        print(result)
        ```
        """
        super().__init__("SWAP", qubits, n_qubits)


class ControlledOperationGate(AbstractGate):
    def __init__(self, gate: str, qubits: ArrayLike, n_qubits: int):
        super().__init__(qubits, n_qubits)
        self.gate = gate
        mat = OPERATIONS_DICT[gate]
        self.register_buffer(
            "matrix",
            create_controlled_matrix_from_operation(
                mat, n_control_qubits=len(qubits) - (int)(math.log2(mat.shape[0]))
            ),
        )

    def matrices(self, _: torch.Tensor) -> torch.Tensor:
        return self.matrix

    def apply(self, matrix: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        return _apply_gate(state, matrix, self.qubits, self.n_qubits)

    def forward(self, state: torch.Tensor, _: torch.Tensor = None) -> torch.Tensor:
        return self.apply(self.matrix, state)


class CNOT(ControlledOperationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a controlled NOT (CNOT) gate in a quantum circuit.

        Arguments:
            qubits (ArrayLike): The control and target qubits for the CNOT gate.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a CNOT gate
        cnot_gate = pyq.CNOT(qubits=[0, 1], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        # Apply the CNOT gate to the zero state
        result = cnot_gate(z_state)
        print(result)
        ```
        """
        super().__init__("X", qubits, n_qubits)


CX = CNOT


class CY(ControlledOperationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a controlled-Y (CY) gate in a quantum circuit.
        The CY Gate class creates a controlled Y gate, applying the Y gate
        according to the control qubit state.

        Arguments:
            qubits (ArrayLike): The control and target qubits for the CY gate.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a CY gate
        cy_gate = pyq.CY(qubits=[0, 1], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        # Apply the CY gate to the zero state
        result = cy_gate(z_state)
        print(result)
        ```
        """
        super().__init__("Y", qubits, n_qubits)


class CZ(ControlledOperationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a controlled-Z (CZ) gate in a quantum circuit.
        The CZ gate class creates a controlled Z gate, applying
        the Z gate according to the control qubit state.


        Arguments:
            qubits (ArrayLike): The control and target qubits for the CZ gate.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a CZ gate
        cz_gate = pyq.CZ(qubits=[0, 1], n_qubits=2)

        # Create a zero state
        z_state = pyq.zero_state(n_qubits=2)

        # Apply the CZ gate to the zero state
        result = cz_gate(z_state)
        print(result)
        ```
        """
        super().__init__("Z", qubits, n_qubits)


class CSWAP(ControlledOperationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a controlled-SWAP (CSWAP) gate in a quantum circuit.
        The CSWAP gate class creates a controlled SWAP gate, applying
        the SWAP gate according to the control qubit state.


        Arguments:
            qubits (ArrayLike): The control and targets qubits for the CSWAP gate.
            n_qubits (int): The total number of qubits in the circuit.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a CSWAP gate
        cswap_gate = pyq.CSWAP(qubits=[0, 1, 2], n_qubits=3)

        # Create a zero state
        swap_state = pyq.zero_state(n_qubits=3)

        # Apply the CSWAP gate to the zero state
        result = cswap_gate(swap_state)
        print(result)
        """
        super().__init__("SWAP", qubits, n_qubits)


class Toffoli(ControlledOperationGate):
    def __init__(self, qubits: ArrayLike, n_qubits: int):
        """
        Represents a multi qubit controlled toffoli gate in a quantum circuit.
        This gate performs a NOT operation only if all the control qubits are in state 1.
        Arguments:
            qubits (ArrayLike): The first n-1 qubits will be considered as the control
                                qubits and the last one will be the target qubit of the
                                Toffoli gate.

        Examples:
        ```python exec="on" source="above" result="json"
        import torch
        import pyqtorch.modules as pyq

        # Create a Toffoli gate with 2 control qubits.
        toffoli_gate = pyq.Toffoli(qubits=[0, 1, 2], n_qubits=3)

        # Create a one state
        state_1111 = pyq.X(qubits=[0], n_qubits=3)(pyq.X(qubits=[1], n_qubits=3)
        (pyq.X(qubits=[2], n_qubits=3)(pyq.zero_state(n_qubits=3)))

        # Apply the Toffoli gate to the zero state
        result = toffoli_gate(toffoli_state)
        print(result)
        """
        super().__init__("X", qubits, n_qubits)
